#ifndef DUNE_SPHEREGRID_GRIDVIEW_HH
#define DUNE_SPHEREGRID_GRIDVIEW_HH

//- dune-common includes
#include <dune/common/typetraits.hh>

//- dune-grid includes
#include <dune/grid/common/capabilities.hh>
#include <dune/grid/common/gridview.hh>

//- dune-metagrid includes
#include <dune/grid/spheregrid/datahandle.hh>
#include <dune/grid/spheregrid/indexset.hh>
#include <dune/grid/spheregrid/intersection.hh>
#include <dune/grid/spheregrid/intersectioniterator.hh>
#include <dune/grid/spheregrid/iterator.hh>

namespace Dune
{

  // Internal Forward Declarations
  // -----------------------------

  template< class HostGridView, class MapToSphere >
  class SphereGridView;



  // SphereGridViewTraits
  // --------------------

  template< class HGV, class MapToSphere >
  class SphereGridViewTraits
  {
    friend class SphereGridView< HGV, MapToSphere >;

    typedef HGV HostGridView;

    typedef typename HostGridView::Grid HostGrid;

  public:
    typedef SphereGridView< HostGridView, MapToSphere > GridViewImp;

    typedef SphereGrid< HostGrid, MapToSphere > Grid;

    typedef SphereGridIndexSet< const Grid, typename HostGridView::IndexSet > IndexSet;

    typedef SphereGridIntersection< const Grid, typename HostGridView::Intersection > IntersectionImpl;
    typedef Dune::Intersection< const Grid, IntersectionImpl > Intersection;

    typedef SphereGridIntersectionIterator< const Grid, typename HostGridView::IntersectionIterator > IntersectionIteratorImpl;
    typedef Dune::IntersectionIterator< const Grid, IntersectionIteratorImpl, IntersectionImpl > IntersectionIterator;

    typedef typename HostGridView::CollectiveCommunication CollectiveCommunication;

    template< int codim >
    struct Codim
    {
      typedef typename Grid::Traits::template Codim< codim >::Entity Entity;

      typedef typename Grid::template Codim< codim >::Geometry Geometry;
      typedef typename Grid::template Codim< codim >::LocalGeometry LocalGeometry;

      template< PartitionIteratorType pitype >
      struct Partition
      {
        typedef typename HostGridView::template Codim< codim >::template Partition< pitype > HostPartition;

        typedef SphereGridIterator< const Grid, typename HostPartition::Iterator > IteratorImpl;
        typedef Dune::EntityIterator< codim, const Grid, IteratorImpl > Iterator;
      };

      typedef typename Partition< All_Partition >::Iterator Iterator;
    };

    static const bool conforming = HostGridView::conforming;
  };



  // SphereGridView
  // --------------

  template< class HGV, class MapToSphere >
  class SphereGridView
  {
    typedef SphereGridView< HGV, MapToSphere > This;

  public:
    typedef SphereGridViewTraits< HGV, MapToSphere > Traits;

    typedef typename Traits::HostGridView HostGridView;

    typedef typename Traits::Grid Grid;

    typedef typename Traits::IndexSet IndexSet;

    typedef typename Traits::Intersection Intersection;

    typedef typename Traits::IntersectionIterator IntersectionIterator;

    typedef typename Traits::CollectiveCommunication CollectiveCommunication;

    template< int codim >
    struct Codim
    : public Traits::template Codim< codim >
    {};

    static const bool conforming = Traits::conforming;

    SphereGridView ( const Grid &grid, const HostGridView &hostGridView )
    : grid_( &grid ),
      hostGridView_( hostGridView )
    {}

    const Grid &grid () const
    {
      assert( grid_ );
      return *grid_;
    }

    const IndexSet &indexSet () const
    {
      if( !indexSet_ )
        indexSet_ = IndexSet( hostGridView().indexSet() );
      return indexSet_;
    }

    int size ( int codim ) const
    {
      return hostGridView().size( codim );
    }

    int size ( const GeometryType &type ) const
    {
      return hostGridView().size( type );
    }

    template< int codim >
    typename Codim< codim >::Iterator begin () const
    {
      return begin< codim, All_Partition >();
    }

    template< int codim, PartitionIteratorType pitype >
    typename Codim< codim >::template Partition< pitype >::Iterator begin () const
    {
      typedef typename Traits::template Codim< codim >::template Partition< pitype >::IteratorImpl Impl;
      return Impl( hostGridView().template begin< codim, pitype >(), mapToSphere() );
    }

    template< int codim >
    typename Codim< codim >::Iterator end () const
    {
      return end< codim, All_Partition >();
    }

    template< int codim, PartitionIteratorType pitype >
    typename Codim< codim >::template Partition< pitype >::Iterator end () const
    {
      typedef typename Traits::template Codim< codim >::template Partition< pitype >::IteratorImpl Impl;
      return Impl( hostGridView().template end< codim, pitype >(), mapToSphere() );
    }

    IntersectionIterator ibegin ( const typename Codim< 0 >::Entity &entity ) const
    {
      typedef typename Traits::IntersectionIteratorImpl IntersectionIteratorImpl;
      return IntersectionIteratorImpl( hostGridView().ibegin( entity.impl().hostEntity() ), mapToSphere() );
    }

    IntersectionIterator iend ( const typename Codim< 0 >::Entity &entity ) const
    {
      typedef typename Traits::IntersectionIteratorImpl IntersectionIteratorImpl;
      return IntersectionIteratorImpl( hostGridView().iend( entity.impl().hostEntity() ), mapToSphere() );
    }

    const CollectiveCommunication &comm () const
    {
      return hostGridView().comm();
    }

    int overlapSize ( int codim ) const
    {
      return hostGridView().overlapSize( codim );
    }

    int ghostSize ( int codim ) const
    {
      return hostGridView().ghostSize( codim );
    }

    template< class DataHandle, class Data >
    void communicate ( CommDataHandleIF< DataHandle, Data > &dataHandle,
                       InterfaceType interface,
                       CommunicationDirection direction ) const
    {
      typedef CommDataHandleIF< DataHandle, Data > DataHandleIF;
      typedef SphereGridDataHandle< DataHandleIF, Grid > WrappedDataHandle;

      WrappedDataHandle wrappedDataHandle( dataHandle, mapToSphere() );
      hostGridView().communicate( wrappedDataHandle, interface, direction );
    }

    const HostGridView &hostGridView () const { return hostGridView_; }

    const MapToSphere &mapToSphere () const { return grid().mapToSphere(); }

  private:
    const Grid *grid_;
    HostGridView hostGridView_;
    mutable IndexSet indexSet_;
  };

} // namespace Dune

#endif // #ifndef DUNE_SPHEREGRID_GRIDVIEW_HH
